#ifndef _UTILS_ENGIN_H_
#define _UTILS_ENGIN_H_

#include <vector>
#include <array>
#include <fstream>
#include <cmath>
#include <algorithm>
#include <Eigen/Dense>
#include <Eigen/CXX11/Tensor>
#include <iostream>



template<typename T>
using Matrix_R = Eigen::Matrix<T, Eigen::Dynamic, Eigen::Dynamic, Eigen::RowMajor>;

template<typename T>
using Matrix_C = Eigen::Matrix<T, Eigen::Dynamic, Eigen::Dynamic, Eigen::ColMajor>;

template<typename T, size_t N>
using Tensor_R = Eigen::Tensor<T, N, Eigen::RowMajor>;

template<typename T, size_t N>
using Tensor_C = Eigen::Tensor<T, N, Eigen::ColMajor>;

template<typename T, size_t dim>
using Arr = std::array<T, dim>;

template<typename T, size_t dim1, size_t dim2>
using Arr2D = std::array<std::array<T, dim2>, dim1>;

template<typename T>
using Vec = std::vector<T>;

template<typename T>
using Vec2D = std::vector<std::vector<T>>;

template<typename T>
using Vec3D = std::vector<std::vector<std::vector<T>>>;


template<typename T>
struct MulticlassRes{
    Matrix_R<T> bboxes;
    Eigen::VectorX<T> scores;
    Eigen::VectorXi labels;
    Eigen::VectorXi dir_scores;
};

template<typename T>
struct PredOut{
    Matrix_R<T> bboxes;
    Eigen::VectorX<T> scores;
    Eigen::VectorXi labels;
};
// ========================= Functions ================== //
template<typename T>
Vec<Tensor_R<T, 4>> create_4D_grid(
    const Eigen::VectorX<T>& x,
    const Eigen::VectorX<T>& y,
    const Eigen::VectorX<T>& z,
    const Eigen::VectorX<T>& r)
{
    Tensor_R<T, 4> X(x.size(), y.size(), z.size(), r.size());
    Tensor_R<T, 4> Y(x.size(), y.size(), z.size(), r.size());
    Tensor_R<T, 4> Z(x.size(), y.size(), z.size(), r.size());
    Tensor_R<T, 4> R(x.size(), y.size(), z.size(), r.size());

    for(int i = 0; i < x.size(); i++)
        for(int j = 0; j < y.size(); j++)
            for(int k = 0; k < z.size(); k++)
                for(int t = 0; t < r.size(); t++){
                    X(i, j, k, t) = x(i);
                    Y(i, j, k, t) = y(j);
                    Z(i, j, k, t) = z(k);
                    R(i, j, k, t) = r(t);
                }
    return Vec<Tensor_R<T, 4>>{X, Y, Z, R};
}

template<typename T, size_t N>
Tensor_R<T, N+1> unsqueeze(Tensor_R<T, N> tensor, size_t dim){
    auto dims = tensor.dimensions();
    std::array<long int, N+1> dims_resh;
    int last_dim = 0;
    for(int i = 0; i < N+1; i++)
        if(i == dim)
            dims_resh[i] = 1;
        else
            dims_resh[i] = dims[last_dim++];

    return tensor.reshape(dims_resh);
}

template<typename T>
inline T sigmoid(T x){
    return 1.0f / (1.0f + std::exp(-x));
}

template<typename T>
Vec<int> topk(const Eigen::VectorX<T>& vec, const size_t k){
    Vec<std::pair<T, int>> val_index_pairs(vec.size());
    for(int i = 0; i < vec.size(); i++)
        val_index_pairs[i] = std::make_pair(vec(i), i);

    std::sort(val_index_pairs.begin(), val_index_pairs.end(),
              [](const std::pair<T, int>& lhs, const std::pair<T, int>& rhs)
              {return lhs.first > rhs.first;});
    Vec<int> indices(k);
    for(int i = 0; i < k; i++)
        indices[i] = val_index_pairs[i].second;

    return indices;
}

template<typename T>
Matrix_R<T> xywhr2xyxyr(const Matrix_R<T>& boxes_xywhr){
    Matrix_R<T> boxes(boxes_xywhr.rows(), boxes_xywhr.cols());
    boxes.setZero();
    Eigen::VectorX<T> half_w = boxes_xywhr.col(2) / 2;
    Eigen::VectorX<T> half_h = boxes_xywhr.col(3) / 2;
    boxes.col(0) = boxes_xywhr.col(0)-half_w;
    boxes.col(1) = boxes_xywhr.col(1)-half_h;
    boxes.col(2) = boxes_xywhr.col(0)+half_w;
    boxes.col(3) = boxes_xywhr.col(1)+half_h;
    boxes.col(4) = boxes_xywhr.col(4);

    return boxes;
}

template<typename T>
Matrix_R<T> cat(const Vec<Matrix_R<T>>& mats){
    auto num_cols = mats[0].cols();
    auto num_rows = mats[0].rows();
    for(int i = 1; i < mats.size(); i++)
        num_rows += mats[i].rows();

    Matrix_R<T> mat_concated(num_rows, num_cols);

    int last_row = 0;
    for(int i = 0; i < mats.size(); i++){
        mat_concated.block(last_row, 0, mats[i].rows(), mats[i].cols()) = mats[i];
        last_row += mats[i].rows();
    }

    return mat_concated;
}

template<typename T>
Eigen::VectorX<T> cat(const Vec<Eigen::VectorX<T>>& vecs){
    auto size = vecs[0].size();
    for(int i = 1; i < vecs.size(); i++)
        size += vecs[i].size();

    Eigen::VectorX<T> vec_concated(size);

    int last_elem = 0;
    for(int i = 0; i < vecs.size(); i++){
        vec_concated.segment(last_elem, vecs[i].size()) = vecs[i];
        last_elem += vecs[i].size();
    }

    return vec_concated;
}

template<typename T>
Eigen::VectorX<T> keep_if(const Eigen::VectorX<T>& vec, const Eigen::VectorX<int> mask){
    auto count = (mask.array() == 1).count();
    Eigen::VectorX<T> vec_kept(count);
    int last_elem = 0;
    for(int i = 0; i < vec.size(); i++)
        if(mask[i] == 1)
            vec_kept[last_elem++] = vec[i];
    return vec_kept;
}

template<typename T>
Matrix_R<T> keep_if(const Matrix_R<T>& mat, const Eigen::VectorX<int> mask){
    auto count = (mask.array() == 1).count();
    Matrix_R<T> mat_kept(count, mat.cols());
    int last_row = 0;
    for(int i = 0; i < mat.rows(); i++)
        if(mask[i] == 1)
            mat_kept.row(last_row++) = mat.row(i);
    return mat_kept;
}

template<typename T>
Eigen::VectorX<T> keep_ind(const Eigen::VectorX<T>& vec, const Vec<int>& ind){
    auto count = ind.size();
    Eigen::VectorX<T> vec_kept(count);
    for(int i = 0; i < count; i++)
        vec_kept[i] = vec[ind[i]];
    return vec_kept;
}

template<typename T>
Matrix_R<T> keep_ind(const Matrix_R<T>& mat, const Vec<int>& ind){
    auto count = ind.size();
    Matrix_R<T> mat_kept(count, mat.cols());
    for(int i = 0; i < ind.size(); i++)
        mat_kept.row(i) = mat.row(ind[i]);
    return mat_kept;
}

template<typename T>
void fill_tensor_from_onnx_output_tensor(const float* vec_ptr, Tensor_R<T, 3>& tensor){
//    std::cout<<"START fill_tensor_from_onnx_output_tensor "<<std::endl;
    auto idx = 0;
    T val;
    auto dim = tensor.dimensions();
//    std::cout<<"after dim fill_tensor_from_onnx_output_tensor "<<std::endl;
    for(auto i = 0; i < dim[0]; i++)
        for(auto j = 0; j < dim[1]; j++)
            for(auto k = 0; k < dim[2]; k++){
                val = vec_ptr[idx];
                tensor(i, j, k) = val;
                idx = idx+1;
            }
}



//template<typename T>
//void fill_tensor_from_file(const std::string& file_name, Tensor_R<T, 3>& tensor){
//    std::ifstream in_file(file_name, std::ios::in);
//    if(!in_file.is_open()){
//        std::cerr << "Unable to open " << file_name << std::endl;
//        return;
//    }
//    T val;
//    auto dim = tensor.dimensions();
//    for(auto i = 0; i < dim[0]; i++)
//        for(auto j = 0; j < dim[1]; j++)
//            for(auto k = 0; k < dim[2]; k++){
//                in_file >> val;
//                tensor(i, j, k) = val;
//            }
//    in_file.close();
//}

//template<typename T>
//void rowwise_maxcoeff_ind(const Matrix_R<T>& mat, Eigen::VectorXi& indices){
//    indices.resize(mat.rows());
//    for(int i = 0; i < mat.rows(); i++)
//        mat.row(i).maxCoeff(&indices[i]);
//}

template<typename T>
void rowwise_maxcoeff_ind(const Eigen::Map<Matrix_R<T>>& mat, Eigen::VectorXi& indices){
    indices.resize(mat.rows());
    for(int i = 0; i < mat.rows(); i++)
        mat.row(i).maxCoeff(&indices[i]);
}

#endif //_UTILS_H_
